#!/usr/bin/python
# greylist -- Courier filter implementing a "greylisting" technique.
# Copyright (C) 2004  Gordon Messmer <gordon@dragonsdawn.net>
# Copyright (C) 2005  Mickael Marchand <marchand@kde.org>
# Copyright (C) 2006  Georg Lutz <georg-list@georglutz.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import anydbm
import md5
import sys
import string
import thread
import time
import courier.control
import re
import whichdb


# should debug messages be written to stderr?
debug = 1

_sendersLock = thread.allocate_lock()
_sendersDir = '/var/state/pythonfilter'
try:
    # Keep a dictionary of sender/recipient pairs that we've seen before
    _dbType = whichdb.whichdb(_sendersDir + '/greylisted')
    _senders = anydbm.open(_sendersDir + '/greylisted', 'c')
    # messages with these mail addresses should not be greylisted (could be your customer database)
    # mail addresses are md5ed for privacy reasons
    _whitelistMailAddresses = anydbm.open(_sendersDir + '/greylist_whitelistMailAddresses', 'c')
    # messages with these sender/receiver domains should not be greylisted
    _whitelistDomains = anydbm.open(_sendersDir + '/greylist_whitelistDomains', 'c')
    # messages from these IP addresses should not be greylisted
    _whitelistIPAddresses = anydbm.open(_sendersDir + '/greylist_whitelistIPAddresses', 'c')
except:
    sys.stderr.write('Failed to open greylist db in %s, make sure that the directory exists\n' % _sendersDir)
    sys.exit(1)



# The good/bad senders lists will be scrubbed at the interval indicated
# in seconds.  All records older than the "listed_for" number of seconds
# will be removed from the lists.
_sendersLastPurged = 0
_sendersPassedTTL = 60 * 60 * 24 * 36
_sendersNotPassedTTL = 60 * 60 * 24 
_sendersPurgeInterval = 60 * 60 * 2
greylistTime = 300


_IPv4Match = re.compile('^(\d+)\.(\d+)\.(\d+)\.(\d+)$')


# Record in the system log that this filter was initialized.
sys.stderr.write('Initialized the "greylist" python filter\n')


def doFilter(bodyFile, controlFileList):
    """Return a temporary failure message if this sender hasn't tried to
    deliver mail previously.

    Search through the control files and discover the envelope sender
    and message recipients.  If the sender has written to these
    recipients before, allow the message.  Otherwise, throw a
    temporary failure, and expect the remote MTA to try again.  Many
    spamware sites and viruses will not, preventing these messages
    from getting into the users' mailbox.

    """

    global _sendersLastPurged
    global greylistTime

    senderIP = courier.control.getSendersIP(controlFileList)
    # check if senderIP is in whitelist
    if senderIP in _whitelistIPAddresses:
	if debug: sys.stderr.write('skipping whitelisted IP address %s\n' % senderIP)
	return ''

    # Calculate the /24 network
    rematch = _IPv4Match.match(senderIP)
    if(rematch == None):
	# IPv6 network calculation not handled yet
	senderIPNetwork = senderIP
    else:
	senderIPNetwork = rematch.group(1) + "." + rematch.group(2) + "." + rematch.group(3)
   
    
    try:
        # The envelope sender will always be the first record in the control
        # files, according to the courier documentation.  Open the first file,
        # read the first line, and if it isn't the sender record, return
        # a failure response.
        ctlfile = open(controlFileList[0])
        ctlline = ctlfile.readline()
    except:
        return '451 Internal failure locating control files'

    if ctlline[0] != 's':
        return '451 Internal failure locating envelope sender record'
    if len(ctlline) == 2:
        # Null sender is allowed as a non-fatal error
        return ''

    sender = string.strip(ctlline[1:])
    sender = string.lower(sender)
    
    tmpSplit = string.split(sender,'@')
    if len(tmpSplit) == 2:
	domainName = tmpSplit[1]
	if domainName in _whitelistDomains:
	    if debug: sys.stderr.write('skipping whitelisted sender domain %s\n' % domainName)
	    return ''


    # Create a new MD5 object.  The pairs of sender/recipient will
    # be stored in the db in the form of an MD5 digest.
    senderMd5 = md5.new(sender)

    # check if sender mail address is whitelisted
    if senderMd5.hexdigest() in _whitelistMailAddresses:
	if debug: sys.stderr.write('skipping whitelisted sender address\n')
	return ''


    # Scrub the lists if it is time to do so.
    _sendersLock.acquire()
    if time.time() > (_sendersLastPurged + _sendersPurgeInterval):
        if debug: sys.stderr.write("Purging _senders list\n")
	minAgePassed = time.time() - _sendersPassedTTL
	minAgeNotPassed = time.time() - _sendersNotPassedTTL
        
	for key in _senders.keys():
            tmpSplit = string.split(_senders[key],"|")
	    firstTimestamp = float(tmpSplit[0])
	    # old format: only first timestamp is recorded
	    if len(tmpSplit) == 1:
		if firstTimestamp < minAgePassed:
		    del _senders[key]
	    # new format: firsttimestamp + "|" + lasttimestamp
	    else:
		lastTimestamp = float(tmpSplit[1])
		# entry has not passed, no second delivery attempt yet
		# --> apply _sendersNotPassedTTL
		if lastTimestamp == firstTimestamp:
		    if firstTimestamp < minAgeNotPassed:
			del _senders[key]
		# entry has already passed / is autowhitelisted
		else:
		    if lastTimestamp < minAgePassed:
			del _senders[key]

        # synchronize the database to disk if the db type supports that
	if (_dbType == 'gdbm') or (_dbType == 'dbhash'):
	    _senders.sync()
	
	_sendersLastPurged = time.time()


   
    

    # Create a digest for each recipient and look it up in the db.
    # Update the timestamp of each pair as we look them up.  If any
    # pair does not exist, we'll have to ask the sender to deliver
    # again.
    foundAll=1
    timeToGo=greylistTime
    biggestTimeToGo=0

    for recipient in courier.control.getRecipients(controlFileList):
        recipient = string.lower(recipient)
        recipientMd5 = md5.new(recipient)


	# check if domainname of receiver mail address is whitelisted
	tmpSplit = string.split(recipient,'@')
	if len(tmpSplit) == 2:
	    domainName = tmpSplit[1]
	    if domainName in _whitelistDomains:
		if debug: sys.stderr.write('skipping whitelisted receiver domain\n')
		_sendersLock.release()
		return ''

	# check if receiver mail address is whitelisted
	if recipientMd5.hexdigest() in _whitelistMailAddresses:
	    if debug: sys.stderr.write('skipping whitelisted receiver address\n')
	    _sendersLock.release()
	    return ''
	

	correspondents = senderMd5.copy()
        correspondents.update(recipient)
	correspondents.update(senderIPNetwork)
        cdigest = correspondents.hexdigest()
        if not _senders.has_key(cdigest):
	    if debug: sys.stderr.write('no greylist entry found\n')
            foundAll = 0
            # firstTimestamp == lastTimestamp for new records
	    _senders[cdigest] = str(time.time()) + "|" + str(time.time())
	    biggestTimeToGo=greylistTime
        else:
	    tmpSplit = string.split(_senders[cdigest],"|")
	    firstTimestamp = float(tmpSplit[0])
	    timeToGo = firstTimestamp + greylistTime - time.time()
            if timeToGo > 0:
		if debug: sys.stderr.write('greylist entry found, but greylist time not reached yet\n')
                foundAll = 0
		if biggestTimeToGo < timeToGo:
		    biggestTimeToGo = timeToGo
	    else:
		# greylist entry found and greylist time exceeded
		if debug: sys.stderr.write('greylist entry found\n')
		# update lastTimestamp
		_senders[cdigest] = str(firstTimestamp) + "|" + str(time.time())
    _sendersLock.release()

    if foundAll:
        return ''
    else:
        return('451 4.7.1 Greylisting in action, please come back in %s' % time.strftime("%H:%M:%S", time.gmtime(biggestTimeToGo)) )


if __name__ == '__main__':
    # For debugging, you can create a file that contains one line,
    # beginning with an 's' character, followed by an email address
    # and more lines, beginning with an 'r' character, for each
    # recipient.  Run this script with the name of that file as an
    # argument, and it'll validate that email address.
    if not sys.argv[1:]:
        print 'Use:  greylist.py <control file>'
        sys.exit(1)
    print doFilter('', sys.argv[1:])
